#!/usr/bin/env python3
"""
GUI wrapper for the OpenInverter CAN Tool
Provides a user-friendly interface for all oic command-line functionality
"""

import tkinter as tk
from tkinter import ttk, messagebox, filedialog, scrolledtext
import subprocess
import threading
import json
import csv
import os
from datetime import datetime
import queue
import re


class OpenInverterGUI:
    def __init__(self, root):
        self.root = root
        self.root.title("OpenInverter CAN Tool GUI")
        self.root.geometry("1000x700")
        
        # Queue for thread communication
        self.output_queue = queue.Queue()
        
        # Variables
        self.database_var = tk.StringVar()
        self.context_var = tk.StringVar()
        self.node_var = tk.StringVar(value="1")
        self.timeout_var = tk.StringVar(value="1.0")
        self.logging_active = False
        self.log_thread = None
        
        # Create GUI
        self.create_widgets()
        self.check_oic_availability()
        
    def create_widgets(self):
        # Main notebook for tabs
        notebook = ttk.Notebook(self.root)
        notebook.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        
        # Connection Settings Tab
        self.create_connection_tab(notebook)
        
        # Parameters Tab
        self.create_parameters_tab(notebook)
        
        # CAN Mapping Tab
        self.create_can_mapping_tab(notebook)
        
        # Commands Tab
        self.create_commands_tab(notebook)
        
        # Logging Tab
        self.create_logging_tab(notebook)
        
        # System Tab
        self.create_system_tab(notebook)
        
    def create_connection_tab(self, notebook):
        frame = ttk.Frame(notebook)
        notebook.add(frame, text="Connection")
        
        # Connection settings
        settings_frame = ttk.LabelFrame(frame, text="Connection Settings")
        settings_frame.pack(fill=tk.X, padx=10, pady=10)
        
        ttk.Label(settings_frame, text="Database File:").grid(row=0, column=0, sticky=tk.W, padx=5, pady=5)
        ttk.Entry(settings_frame, textvariable=self.database_var, width=50).grid(row=0, column=1, padx=5, pady=5)
        ttk.Button(settings_frame, text="Browse", command=self.browse_database).grid(row=0, column=2, padx=5, pady=5)
        
        ttk.Label(settings_frame, text="Context:").grid(row=1, column=0, sticky=tk.W, padx=5, pady=5)
        ttk.Entry(settings_frame, textvariable=self.context_var, width=20).grid(row=1, column=1, sticky=tk.W, padx=5, pady=5)
        
        ttk.Label(settings_frame, text="Node ID:").grid(row=2, column=0, sticky=tk.W, padx=5, pady=5)
        ttk.Entry(settings_frame, textvariable=self.node_var, width=10).grid(row=2, column=1, sticky=tk.W, padx=5, pady=5)
        
        ttk.Label(settings_frame, text="Timeout (s):").grid(row=3, column=0, sticky=tk.W, padx=5, pady=5)
        ttk.Entry(settings_frame, textvariable=self.timeout_var, width=10).grid(row=3, column=1, sticky=tk.W, padx=5, pady=5)
        
        # Test connection
        test_frame = ttk.LabelFrame(frame, text="Connection Test")
        test_frame.pack(fill=tk.X, padx=10, pady=10)
        
        ttk.Button(test_frame, text="Test Connection", command=self.test_connection).pack(pady=10)
        ttk.Button(test_frame, text="Scan CAN Bus", command=self.scan_can_bus).pack(pady=5)
        ttk.Button(test_frame, text="Get Serial Number", command=self.get_serial_number).pack(pady=5)
        
    def create_parameters_tab(self, notebook):
        frame = ttk.Frame(notebook)
        notebook.add(frame, text="Parameters")
        
        # Parameter operations
        param_frame = ttk.LabelFrame(frame, text="Parameter Operations")
        param_frame.pack(fill=tk.X, padx=10, pady=10)
        
        # List parameters
        ttk.Button(param_frame, text="List All Parameters", command=self.list_parameters).pack(pady=5)
        
        # Read parameter
        read_frame = ttk.Frame(param_frame)
        read_frame.pack(fill=tk.X, pady=5)
        ttk.Label(read_frame, text="Read Parameter:").pack(side=tk.LEFT)
        self.read_param_var = tk.StringVar()
        ttk.Entry(read_frame, textvariable=self.read_param_var, width=30).pack(side=tk.LEFT, padx=5)
        ttk.Button(read_frame, text="Read", command=self.read_parameter).pack(side=tk.LEFT, padx=5)
        
        # Write parameter
        write_frame = ttk.Frame(param_frame)
        write_frame.pack(fill=tk.X, pady=5)
        ttk.Label(write_frame, text="Write Parameter:").pack(side=tk.LEFT)
        self.write_param_var = tk.StringVar()
        self.write_value_var = tk.StringVar()
        ttk.Entry(write_frame, textvariable=self.write_param_var, width=20).pack(side=tk.LEFT, padx=5)
        ttk.Label(write_frame, text="Value:").pack(side=tk.LEFT)
        ttk.Entry(write_frame, textvariable=self.write_value_var, width=20).pack(side=tk.LEFT, padx=5)
        ttk.Button(write_frame, text="Write", command=self.write_parameter).pack(side=tk.LEFT, padx=5)
        
        # Dump all parameters
        ttk.Button(param_frame, text="Dump All Parameters", command=self.dump_all_parameters).pack(pady=5)
        
        # Save/Load parameters
        file_frame = ttk.Frame(param_frame)
        file_frame.pack(fill=tk.X, pady=5)
        ttk.Button(file_frame, text="Save Parameters to JSON", command=self.save_parameters).pack(side=tk.LEFT, padx=5)
        ttk.Button(file_frame, text="Load Parameters from JSON", command=self.load_parameters).pack(side=tk.LEFT, padx=5)
        
    def create_can_mapping_tab(self, notebook):
        frame = ttk.Frame(notebook)
        notebook.add(frame, text="CAN Mapping")
        
        # CAN mapping operations
        can_frame = ttk.LabelFrame(frame, text="CAN Message Mapping")
        can_frame.pack(fill=tk.X, padx=10, pady=10)
        
        ttk.Button(can_frame, text="List CAN Mappings", command=self.list_can_mappings).pack(pady=5)
        
        # Create mapping
        create_frame = ttk.Frame(can_frame)
        create_frame.pack(fill=tk.X, pady=5)
        ttk.Label(create_frame, text="Create Mapping - Parameter:").pack(side=tk.LEFT)
        self.can_param_var = tk.StringVar()
        ttk.Entry(create_frame, textvariable=self.can_param_var, width=20).pack(side=tk.LEFT, padx=5)
        ttk.Label(create_frame, text="CAN ID:").pack(side=tk.LEFT)
        self.can_id_var = tk.StringVar()
        ttk.Entry(create_frame, textvariable=self.can_id_var, width=10).pack(side=tk.LEFT, padx=5)
        ttk.Label(create_frame, text="Position:").pack(side=tk.LEFT)
        self.can_pos_var = tk.StringVar()
        ttk.Entry(create_frame, textvariable=self.can_pos_var, width=5).pack(side=tk.LEFT, padx=5)
        ttk.Button(create_frame, text="Create", command=self.create_can_mapping).pack(side=tk.LEFT, padx=5)
        
        # Remove mapping
        remove_frame = ttk.Frame(can_frame)
        remove_frame.pack(fill=tk.X, pady=5)
        ttk.Label(remove_frame, text="Remove Mapping Index:").pack(side=tk.LEFT)
        self.can_remove_var = tk.StringVar()
        ttk.Entry(remove_frame, textvariable=self.can_remove_var, width=10).pack(side=tk.LEFT, padx=5)
        ttk.Button(remove_frame, text="Remove", command=self.remove_can_mapping).pack(side=tk.LEFT, padx=5)
        
        # Export/Import mappings
        export_frame = ttk.Frame(can_frame)
        export_frame.pack(fill=tk.X, pady=5)
        ttk.Button(export_frame, text="Export Mappings to JSON", command=self.export_can_mappings).pack(side=tk.LEFT, padx=5)
        ttk.Button(export_frame, text="Import Mappings from JSON", command=self.import_can_mappings).pack(side=tk.LEFT, padx=5)
        ttk.Button(export_frame, text="Export to DBC", command=self.export_dbc).pack(side=tk.LEFT, padx=5)
        
    def create_commands_tab(self, notebook):
        frame = ttk.Frame(notebook)
        notebook.add(frame, text="Commands")
        
        # Device commands
        cmd_frame = ttk.LabelFrame(frame, text="Device Commands")
        cmd_frame.pack(fill=tk.X, padx=10, pady=10)
        
        # Arrange buttons in a grid
        cmd_buttons = [
            ("Start Inverter", self.start_inverter),
            ("Stop Inverter", self.stop_inverter),
            ("Load Parameters from Flash", self.load_params_flash),
            ("Save Parameters to Flash", self.save_params_flash),
            ("Revert to Defaults", self.revert_defaults),
            ("Reset Device", self.reset_device)
        ]
        
        for i, (text, command) in enumerate(cmd_buttons):
            row = i // 2
            col = i % 2
            ttk.Button(cmd_frame, text=text, command=command).grid(row=row, column=col, padx=10, pady=5, sticky=tk.W+tk.E)
        
        # Configure grid weights
        cmd_frame.columnconfigure(0, weight=1)
        cmd_frame.columnconfigure(1, weight=1)
        
    def create_logging_tab(self, notebook):
        frame = ttk.Frame(notebook)
        notebook.add(frame, text="Logging")
        
        # Logging setup
        log_frame = ttk.LabelFrame(frame, text="Parameter Logging")
        log_frame.pack(fill=tk.X, padx=10, pady=10)
        
        # Parameters to log
        param_frame = ttk.Frame(log_frame)
        param_frame.pack(fill=tk.X, pady=5)
        ttk.Label(param_frame, text="Parameters to log (comma-separated):").pack(anchor=tk.W)
        self.log_params_var = tk.StringVar(value="udc,idc,temp,speed")
        ttk.Entry(param_frame, textvariable=self.log_params_var, width=60).pack(fill=tk.X, pady=5)
        
        # Log interval
        interval_frame = ttk.Frame(log_frame)
        interval_frame.pack(fill=tk.X, pady=5)
        ttk.Label(interval_frame, text="Log interval (seconds):").pack(side=tk.LEFT)
        self.log_interval_var = tk.StringVar(value="1")
        ttk.Entry(interval_frame, textvariable=self.log_interval_var, width=10).pack(side=tk.LEFT, padx=5)
        
        # Output file
        file_frame = ttk.Frame(log_frame)
        file_frame.pack(fill=tk.X, pady=5)
        ttk.Label(file_frame, text="Output file:").pack(side=tk.LEFT)
        self.log_file_var = tk.StringVar()
        ttk.Entry(file_frame, textvariable=self.log_file_var, width=40).pack(side=tk.LEFT, padx=5)
        ttk.Button(file_frame, text="Browse", command=self.browse_log_file).pack(side=tk.LEFT, padx=5)
        
        # Start/Stop logging
        control_frame = ttk.Frame(log_frame)
        control_frame.pack(fill=tk.X, pady=10)
        self.start_log_btn = ttk.Button(control_frame, text="Start Logging", command=self.start_logging)
        self.start_log_btn.pack(side=tk.LEFT, padx=5)
        self.stop_log_btn = ttk.Button(control_frame, text="Stop Logging", command=self.stop_logging, state=tk.DISABLED)
        self.stop_log_btn.pack(side=tk.LEFT, padx=5)
        
        # Log status
        self.log_status_var = tk.StringVar(value="Logging stopped")
        ttk.Label(control_frame, textvariable=self.log_status_var).pack(side=tk.LEFT, padx=20)
        
    def create_system_tab(self, notebook):
        frame = ttk.Frame(notebook)
        notebook.add(frame, text="System")
        
        # Firmware upgrade
        fw_frame = ttk.LabelFrame(frame, text="Firmware Management")
        fw_frame.pack(fill=tk.X, padx=10, pady=10)
        
        upgrade_frame = ttk.Frame(fw_frame)
        upgrade_frame.pack(fill=tk.X, pady=5)
        ttk.Label(upgrade_frame, text="Firmware file:").pack(side=tk.LEFT)
        self.fw_file_var = tk.StringVar()
        ttk.Entry(upgrade_frame, textvariable=self.fw_file_var, width=40).pack(side=tk.LEFT, padx=5)
        ttk.Button(upgrade_frame, text="Browse", command=self.browse_firmware).pack(side=tk.LEFT, padx=5)
        ttk.Button(upgrade_frame, text="Upgrade", command=self.upgrade_firmware).pack(side=tk.LEFT, padx=5)
        
        # Cache management
        cache_frame = ttk.LabelFrame(frame, text="Cache Management")
        cache_frame.pack(fill=tk.X, padx=10, pady=10)
        
        ttk.Button(cache_frame, text="Clear Cache", command=self.clear_cache).pack(pady=5)
        ttk.Button(cache_frame, text="Show Cache Info", command=self.show_cache_info).pack(pady=5)
        
        # Output area
        output_frame = ttk.LabelFrame(frame, text="Command Output")
        output_frame.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        
        self.output_text = scrolledtext.ScrolledText(output_frame, height=15, width=80)
        self.output_text.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        # Clear output button
        ttk.Button(output_frame, text="Clear Output", command=self.clear_output).pack(pady=5)
        
    def check_oic_availability(self):
        """Check if oic command is available"""
        try:
            result = subprocess.run(['oic', '--version'], capture_output=True, text=True, timeout=5)
            if result.returncode == 0:
                self.log_output(f"OpenInverter CAN Tool found: {result.stdout.strip()}")
            else:
                self.log_output("Warning: oic command not found. Please install openinverter-can-tool")
        except (subprocess.TimeoutExpired, FileNotFoundError):
            messagebox.showwarning("Warning", 
                                 "OpenInverter CAN Tool (oic) not found!\n\n"
                                 "Please install it with:\n"
                                 "pip install openinverter-can-tool")
    
    def get_base_command(self):
        """Build base oic command with common options"""
        cmd = ['oic']
        
        if self.database_var.get():
            cmd.extend(['-d', self.database_var.get()])
        if self.context_var.get():
            cmd.extend(['-c', self.context_var.get()])
        if self.node_var.get():
            cmd.extend(['-n', self.node_var.get()])
        if self.timeout_var.get():
            cmd.extend(['-t', self.timeout_var.get()])
            
        return cmd
    
    def run_command(self, cmd_args, input_text=None):
        """Run oic command in a separate thread"""
        def run():
            try:
                cmd = self.get_base_command() + cmd_args
                self.log_output(f"Running: {' '.join(cmd)}")
                
                result = subprocess.run(cmd, capture_output=True, text=True, 
                                      input=input_text, timeout=30)
                
                if result.stdout:
                    self.log_output(result.stdout)
                if result.stderr:
                    self.log_output(f"Error: {result.stderr}")
                if result.returncode != 0:
                    self.log_output(f"Command failed with return code {result.returncode}")
                    
            except subprocess.TimeoutExpired:
                self.log_output("Error: Command timed out")
            except Exception as e:
                self.log_output(f"Error: {str(e)}")
                
        thread = threading.Thread(target=run)
        thread.daemon = True
        thread.start()
    
    def log_output(self, text):
        """Add text to output area"""
        self.output_text.insert(tk.END, f"[{datetime.now().strftime('%H:%M:%S')}] {text}\n")
        self.output_text.see(tk.END)
        self.root.update_idletasks()
    
    def clear_output(self):
        """Clear the output text area"""
        self.output_text.delete(1.0, tk.END)
    
    # File browser methods
    def browse_database(self):
        filename = filedialog.askopenfilename(
            title="Select Parameter Database",
            filetypes=[("JSON files", "*.json"), ("All files", "*.*")]
        )
        if filename:
            self.database_var.set(filename)
    
    def browse_log_file(self):
        filename = filedialog.asksaveasfilename(
            title="Select Log File",
            defaultextension=".csv",
            filetypes=[("CSV files", "*.csv"), ("All files", "*.*")]
        )
        if filename:
            self.log_file_var.set(filename)
    
    def browse_firmware(self):
        filename = filedialog.askopenfilename(
            title="Select Firmware File",
            filetypes=[("Binary files", "*.bin"), ("All files", "*.*")]
        )
        if filename:
            self.fw_file_var.set(filename)
    
    # Connection methods
    def test_connection(self):
        self.run_command(['serialno'])
    
    def scan_can_bus(self):
        self.run_command(['scan'])
    
    def get_serial_number(self):
        self.run_command(['serialno'])
    
    # Parameter methods
    def list_parameters(self):
        self.run_command(['listparams'])
    
    def read_parameter(self):
        param = self.read_param_var.get().strip()
        if not param:
            messagebox.showwarning("Warning", "Please enter a parameter name")
            return
        self.run_command(['read', param])
    
    def write_parameter(self):
        param = self.write_param_var.get().strip()
        value = self.write_value_var.get().strip()
        if not param or not value:
            messagebox.showwarning("Warning", "Please enter both parameter name and value")
            return
        self.run_command(['write', param, value])
    
    def dump_all_parameters(self):
        self.run_command(['dumpall'])
    
    def save_parameters(self):
        filename = filedialog.asksaveasfilename(
            title="Save Parameters",
            defaultextension=".json",
            filetypes=[("JSON files", "*.json"), ("All files", "*.*")]
        )
        if filename:
            self.run_command(['save', filename])
    
    def load_parameters(self):
        filename = filedialog.askopenfilename(
            title="Load Parameters",
            filetypes=[("JSON files", "*.json"), ("All files", "*.*")]
        )
        if filename:
            self.run_command(['load', filename])
    
    # CAN mapping methods
    def list_can_mappings(self):
        self.run_command(['can', 'list'])
    
    def create_can_mapping(self):
        param = self.can_param_var.get().strip()
        can_id = self.can_id_var.get().strip()
        position = self.can_pos_var.get().strip()
        
        if not all([param, can_id, position]):
            messagebox.showwarning("Warning", "Please fill in all fields")
            return
            
        self.run_command(['can', 'create', param, can_id, position])
    
    def remove_can_mapping(self):
        index = self.can_remove_var.get().strip()
        if not index:
            messagebox.showwarning("Warning", "Please enter mapping index")
            return
        self.run_command(['can', 'remove', index])
    
    def export_can_mappings(self):
        filename = filedialog.asksaveasfilename(
            title="Export CAN Mappings",
            defaultextension=".json",
            filetypes=[("JSON files", "*.json"), ("All files", "*.*")]
        )
        if filename:
            self.run_command(['can', 'export', filename])
    
    def import_can_mappings(self):
        filename = filedialog.askopenfilename(
            title="Import CAN Mappings",
            filetypes=[("JSON files", "*.json"), ("All files", "*.*")]
        )
        if filename:
            self.run_command(['can', 'import', filename])
    
    def export_dbc(self):
        filename = filedialog.asksaveasfilename(
            title="Export to DBC",
            defaultextension=".dbc",
            filetypes=[("DBC files", "*.dbc"), ("All files", "*.*")]
        )
        if filename:
            self.run_command(['can', 'dbc', filename])
    
    # Command methods
    def start_inverter(self):
        self.run_command(['cmd', 'start'])
    
    def stop_inverter(self):
        self.run_command(['cmd', 'stop'])
    
    def load_params_flash(self):
        self.run_command(['cmd', 'load'])
    
    def save_params_flash(self):
        self.run_command(['cmd', 'save'])
    
    def revert_defaults(self):
        if messagebox.askyesno("Confirm", "Are you sure you want to revert all parameters to defaults?"):
            self.run_command(['cmd', 'defaults'])
    
    def reset_device(self):
        if messagebox.askyesno("Confirm", "Are you sure you want to reset the device?"):
            self.run_command(['cmd', 'reset'])
    
    # Logging methods
    def start_logging(self):
        params = self.log_params_var.get().strip()
        interval = self.log_interval_var.get().strip()
        filename = self.log_file_var.get().strip()
        
        if not all([params, interval, filename]):
            messagebox.showwarning("Warning", "Please fill in all logging fields")
            return
        
        if self.logging_active:
            messagebox.showwarning("Warning", "Logging is already active")
            return
        
        # Start logging in a separate thread
        def log_worker():
            try:
                cmd = self.get_base_command() + ['log', '-i', interval, '-o', filename] + params.split(',')
                self.log_output(f"Starting logging: {' '.join(cmd)}")
                
                self.log_process = subprocess.Popen(cmd, stdout=subprocess.PIPE, 
                                                  stderr=subprocess.PIPE, text=True)
                
                self.logging_active = True
                self.root.after(0, self.update_log_ui, True)
                
                # Monitor the process
                while self.logging_active and self.log_process.poll() is None:
                    try:
                        line = self.log_process.stdout.readline()
                        if line:
                            self.root.after(0, self.log_output, f"LOG: {line.strip()}")
                    except:
                        break
                
                if self.log_process.returncode and self.log_process.returncode != 0:
                    stderr = self.log_process.stderr.read()
                    if stderr:
                        self.root.after(0, self.log_output, f"Logging error: {stderr}")
                
            except Exception as e:
                self.root.after(0, self.log_output, f"Logging error: {str(e)}")
            finally:
                self.logging_active = False
                self.root.after(0, self.update_log_ui, False)
        
        self.log_thread = threading.Thread(target=log_worker)
        self.log_thread.daemon = True
        self.log_thread.start()
    
    def stop_logging(self):
        if hasattr(self, 'log_process') and self.log_process.poll() is None:
            self.log_process.terminate()
        self.logging_active = False
        self.update_log_ui(False)
        self.log_output("Logging stopped")
    
    def update_log_ui(self, logging_active):
        if logging_active:
            self.start_log_btn.config(state=tk.DISABLED)
            self.stop_log_btn.config(state=tk.NORMAL)
            self.log_status_var.set("Logging active")
        else:
            self.start_log_btn.config(state=tk.NORMAL)
            self.stop_log_btn.config(state=tk.DISABLED)
            self.log_status_var.set("Logging stopped")
    
    # System methods
    def upgrade_firmware(self):
        filename = self.fw_file_var.get().strip()
        if not filename:
            messagebox.showwarning("Warning", "Please select a firmware file")
            return
        
        if messagebox.askyesno("Confirm", "Are you sure you want to upgrade firmware?\n"
                                        "This operation may take several minutes."):
            self.run_command(['upgrade', filename])
    
    def clear_cache(self):
        self.run_command(['cache', 'clear'])
    
    def show_cache_info(self):
        self.run_command(['cache', 'info'])


def main():
    root = tk.Tk()
    app = OpenInverterGUI(root)
    
    # Handle window closing
    def on_closing():
        if hasattr(app, 'logging_active') and app.logging_active:
            app.stop_logging()
        root.destroy()
    
    root.protocol("WM_DELETE_WINDOW", on_closing)
    root.mainloop()


if __name__ == "__main__":
    main()